package de.fraunhofer.iem.dataprovider.kpi.service

import de.fraunhofer.iem.dataprovider.kpi.dto.KPITreeChildResponseDto
import de.fraunhofer.iem.dataprovider.kpi.dto.KPITreeResponseDto
import de.fraunhofer.iem.dataprovider.kpi.dto.KpiCalculationDto
import de.fraunhofer.iem.dataprovider.kpi.dto.RawValueKpiCreateDto
import de.fraunhofer.iem.dataprovider.kpi.entity.KPIEntity
import de.fraunhofer.iem.dataprovider.kpi.enumeration.KpiKind
import de.fraunhofer.iem.dataprovider.kpi.repository.KPIRepository
import de.fraunhofer.iem.dataprovider.kpi.strategy.AggregationKPICalculationStrategy
import de.fraunhofer.iem.dataprovider.kpi.strategy.MaximumKPICalculationStrategy
import de.fraunhofer.iem.dataprovider.kpi.strategy.RatioKPICalculationStrategy
import de.fraunhofer.iem.dataprovider.logger.getLogger
import de.fraunhofer.iem.dataprovider.repository.dto.RepositoryDetailsDto
import de.fraunhofer.iem.dataprovider.toolRun.entity.ToolRunEntity
import de.fraunhofer.iem.dataprovider.tools.occmd.enumeration.Checks
import de.fraunhofer.iem.dataprovider.tools.occmd.json.RawResultJson
import de.fraunhofer.iem.dataprovider.tools.ort.dto.VulnerabilityDto
import org.springframework.stereotype.Service

@Service
class KPIService(
    private val kpiRepository: KPIRepository,
) {

    private val logger = getLogger(javaClass)

    fun removeKPIChildrenLowerThanSecondLevel(kpiTreeResponseDto: KPITreeResponseDto): KPITreeResponseDto {
        val children: List<KPITreeChildResponseDto> = kpiTreeResponseDto.children.map {
            KPITreeChildResponseDto(
                kpi = KPITreeResponseDto(
                    value = it.kpi.value,
                    name = it.kpi.name,
                    description = it.kpi.description,
                    isRoot = it.kpi.isRoot,
                    displayValue = it.kpi.displayValue,
                    emptyList(),
                    it.kpi.isEmpty,
                    order = it.kpi.order
                ),
                actualWeight = it.actualWeight,
                plannedWeight = it.plannedWeight
            )
        }

        return KPITreeResponseDto(
            value = kpiTreeResponseDto.value,
            name = kpiTreeResponseDto.name,
            description = kpiTreeResponseDto.description,
            isRoot = true,
            displayValue = kpiTreeResponseDto.displayValue,
            children = children,
            isEmpty = kpiTreeResponseDto.isEmpty,
            order = kpiTreeResponseDto.order
        )
    }

    /**
     * This method calculates the OccmdCreateKpiDtos based upon the given tool results.
     */
    fun calculateOccmdKpis(rawOccmdResults: List<RawResultJson>): List<RawValueKpiCreateDto> {
        val kpis = mutableListOf<RawValueKpiCreateDto>()

        rawOccmdResults.forEach {
            when (val check = Checks.fromString(it.check)) {
                Checks.CheckedInBinaries ->
                    kpis.add(
                        RawValueKpiCreateDto(
                            kind = KpiKind.CHECKED_IN_BINARIES,
                            score = (it.score * 100).toInt()
                        )
                    )

                Checks.SastUsageBasic ->
                    kpis.add(
                        RawValueKpiCreateDto(
                            kind = KpiKind.SAST_USAGE,
                            score = (it.score * 100).toInt()
                        )
                    )

                Checks.Secrets ->
                    kpis.add(
                        RawValueKpiCreateDto(
                            kind = KpiKind.SECRETS,
                            score = (it.score * 100).toInt()
                        )
                    )

                Checks.CommentsInCode ->
                    kpis.add(
                        RawValueKpiCreateDto(
                            kind = KpiKind.COMMENTS_IN_CODE,
                            score = (it.score * 100).toInt()
                        )
                    )

                Checks.DocumentationInfrastructure ->
                    kpis.add(
                        RawValueKpiCreateDto(
                            kind = KpiKind.DOCUMENTATION_INFRASTRUCTURE,
                            score = (it.score * 100).toInt()
                        )
                    )

                else ->
                    logger.warn("Unknown OCCMD result. $check")
            }
        }

        return kpis
    }

    fun calculateVulnerabilityKpis(vulnerabilityDtos: List<VulnerabilityDto>): List<RawValueKpiCreateDto> {
        return vulnerabilityDtos.map {
            RawValueKpiCreateDto(
                kind = KpiKind.VULNERABILITY_SCORE,
                score = (it.severity * 10).toInt()
            )
        }
    }

    /**
     * Creates a named map of RepositoryCreateDtos, based upon the provided repository details.
     * This method only returns raw KPIs.
     */
    fun calculateRepositoryDetailsKpis(repoDetailsDto: RepositoryDetailsDto): List<RawValueKpiCreateDto> {
        return listOf(
            RawValueKpiCreateDto(
                kind = KpiKind.NUMBER_OF_COMMITS,
                score = repoDetailsDto.numberOfCommits
            ),
            RawValueKpiCreateDto(
                kind = KpiKind.NUMBER_OF_SIGNED_COMMITS,
                score = repoDetailsDto.numberOfSignedCommits
            ),
            RawValueKpiCreateDto(
                kind = KpiKind.IS_DEFAULT_BRANCH_PROTECTED,
                score = if (repoDetailsDto.isDefaultBranchProtected) 100 else 0
            )
        )
    }

    fun saveRawKpis(
        toolRunEntity: ToolRunEntity,
        rawKpis: List<RawValueKpiCreateDto>
    ): List<KPIEntity> {
        return kpiRepository.saveAll(rawKpis.map { it.toDbObject(toolRunEntity) })
    }

    fun getKpiTreeForRawKpis(rawKpis: List<KpiCalculationDto>): KpiCalculationDto {
        val rawValueKpis = rawKpis.toMutableList()
        val vulnerabilities = rawValueKpis.filter { it.kind == KpiKind.VULNERABILITY_SCORE }
        rawValueKpis.toMutableList().removeAll(vulnerabilities)
        val kpiMap = rawValueKpis.associateBy { it.kind }
        val rootKpi = generateKPITree(kpiMap, vulnerabilities)
        calculateKPIsRecursively(rootKpi)
        return rootKpi
    }

    @Suppress("MagicNumber", "LongMethod")
    private fun generateKPITree(
        rawValueKpis: Map<KpiKind, KpiCalculationDto> = emptyMap(),
        vulnerabilityKpis: List<KpiCalculationDto> = emptyList()
    ): KpiCalculationDto {
        val signedCommitsRatioKPI =
            KpiCalculationDto(
                kind = KpiKind.SIGNED_COMMITS_RATIO,
                calculationStrategy = RatioKPICalculationStrategy()
            )

        signedCommitsRatioKPI.addChildKPI(
            rawValueKpis.getOrDefault(
                KpiKind.NUMBER_OF_COMMITS,
                KpiCalculationDto(kind = KpiKind.NUMBER_OF_COMMITS)
            ),
            1.0
        )

        signedCommitsRatioKPI.addChildKPI(
            rawValueKpis.getOrDefault(
                KpiKind.NUMBER_OF_SIGNED_COMMITS,
                KpiCalculationDto(kind = KpiKind.NUMBER_OF_SIGNED_COMMITS)
            ),
            1.0
        )

        val processComplianceKPI = KpiCalculationDto(
            kind = KpiKind.PROCESS_COMPLIANCE,
            calculationStrategy = AggregationKPICalculationStrategy()
        )

        processComplianceKPI.addChildKPI(
            rawValueKpis.getOrDefault(
                KpiKind.IS_DEFAULT_BRANCH_PROTECTED,
                KpiCalculationDto(kind = KpiKind.IS_DEFAULT_BRANCH_PROTECTED)
            ),
            0.5
        )
        processComplianceKPI.addChildKPI(signedCommitsRatioKPI, 0.5)

        val documentationKpi = KpiCalculationDto(
            kind = KpiKind.DOCUMENTATION,
            calculationStrategy = AggregationKPICalculationStrategy()
        )

        documentationKpi.addChildKPI(
            rawValueKpis.getOrDefault(
                KpiKind.DOCUMENTATION_INFRASTRUCTURE,
                KpiCalculationDto(kind = KpiKind.DOCUMENTATION_INFRASTRUCTURE)
            ),
            0.6
        )

        documentationKpi.addChildKPI(
            rawValueKpis.getOrDefault(
                KpiKind.COMMENTS_IN_CODE,
                KpiCalculationDto(kind = KpiKind.COMMENTS_IN_CODE)
            ),
            0.4
        )

        val processTransparencyKPI = KpiCalculationDto(
            kind = KpiKind.PROCESS_TRANSPARENCY,
            calculationStrategy = AggregationKPICalculationStrategy()
        )
        processTransparencyKPI.addChildKPI(documentationKpi, 0.5)
        processTransparencyKPI.addChildKPI(signedCommitsRatioKPI, 0.5)

        val securityKPI = KpiCalculationDto(
            kind = KpiKind.SECURITY,
            calculationStrategy = AggregationKPICalculationStrategy()
        )

        val maximalDependencyVulnerabilityKPI = KpiCalculationDto(
            kind = KpiKind.MAXIMAL_VULNERABILITY,
            calculationStrategy = MaximumKPICalculationStrategy()
        )

        vulnerabilityKpis.forEach { maximalDependencyVulnerabilityKPI.addChildKPI(it, 0.0) }

        securityKPI.addChildKPI(
            rawValueKpis.getOrDefault(
                KpiKind.SECRETS,
                KpiCalculationDto(kind = KpiKind.SECRETS)
            ),
            0.3
        )

        securityKPI.addChildKPI(
            rawValueKpis.getOrDefault(
                KpiKind.CHECKED_IN_BINARIES,
                KpiCalculationDto(kind = KpiKind.CHECKED_IN_BINARIES)
            ),
            0.2
        )

        securityKPI.addChildKPI(maximalDependencyVulnerabilityKPI, 0.5)

        val rootKPI = KpiCalculationDto(
            kind = KpiKind.ROOT,
            calculationStrategy = AggregationKPICalculationStrategy(),
        )

        rootKPI.addChildKPI(processTransparencyKPI, 0.25)
        rootKPI.addChildKPI(processComplianceKPI, 0.25)
        rootKPI.addChildKPI(securityKPI, 0.5)
        return rootKPI
    }

    private fun calculateKPIsRecursively(
        kpi: KpiCalculationDto,
        visited: MutableSet<KpiCalculationDto> = mutableSetOf()
    ) {
        // Check if the KPI has already been visited
        if (visited.contains(kpi)) {
            return
        }

        // Check if the KPI has child KPIs
        if (!kpi.hasChildren()) {
            // Leaf node, calculate the KPI value
            kpi.calculateKPI()
            visited.add(kpi)
            return
        }

        // Recursively calculate child KPIs first
        for (childEdge in kpi.getHierarchyEdges()) {
            calculateKPIsRecursively(childEdge.to, visited)
        }

        // Calculate the KPI value after processing child KPIs
        kpi.calculateKPI()
        visited.add(kpi)
    }
}
